#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 09:28:06 2020

@author: andrewthomson
"""

#%% IMPORT REQUIRED LIBRARIES
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import seaborn as sns

from sklearn.mixture import GaussianMixture
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn import model_selection
scaler = StandardScaler()

import math
import time
import pickle #for python 3.8
#import pickle5 as pickle #for python 3.7 or below

timestr = time.strftime("%Y%m%d-%H%M%S")
timestrb = time.strftime("%Y%m%d")

#%% IMPORT LITERATURE DATA
lit = pd.read_csv('lit_maj_nat_030720.txt',sep="\t") # LITERATURE MAJORITE DATASET

#CUT OUT NON_MAJOIRITE SAMPLES BASED ON NAMJ+MAJ COMPONENTs
namj = np.empty([len(lit)])
maj = np.empty([len(lit)])
tot_maj = np.empty([len(lit)])
added = np.empty([len(lit)])

for i in range(0,len(lit)):
    if (lit.Na[i]-lit.Ti[i])/2 >= 0:
        namj[i] = (lit.Na[i]-lit.Ti[i])/2
    else:
        namj[i] = 0

    if (lit.Si[i]-(2*namj[i])-3) >= 0:
        maj[i] = (lit.Si[i]-(2*namj[i])-3)
    else:
        maj[i] = 0
        
    tot_maj[i] = namj[i]+maj[i]
    
    added[i] = lit.Si[i] + lit.Ti[i] + lit.Al[i] + lit.Cr[i] + lit.Fet[i] + lit.Mn[i] + lit.Mg[i] + lit.Ca[i] + lit.Na[i]
    
lit['tot_maj'] = tot_maj
lit['added'] = added
lit2 = lit[lit['tot_maj'] >= 0.005]
lit3 = lit2[lit2['added'] >= 7.9]
lit4 = lit3[lit3['added'] <= 8.1]

del lit2, lit3, added, tot_maj, namj, maj
lit4 = lit4.reset_index(drop=True)
lit5 = lit4.reset_index(drop=True)
df = lit4.drop(columns=['tot_maj','ExptT','ref','added','mark','col']) # LITERATURE MAJORITE DATASET - only pressure and predictor variables

df.to_csv(r'data_filteredforvalidation.txt',sep='\t',index=False,header=True)

#%% SET UP RANDOM FOREST REGRESSION LOOP

target_column = ['ExptP'] # SET PRESSURE AS TARGET VARIABLE
predictors = ["Si", "Na", "Al", "Mg", "Fet", "Ca", "Mn", "Cr", "Ti"]

# fraction of input data to use as test dataset (remainder is training datset) per MC loop
frac=0.3
# number of MC repetitions
repeat = 1000

# Create arrays that will be populated in RFR MC loop
nopredict = len(predictors)
roundno = math.ceil(len(lit4)*frac)
roundnob = len(lit4) - roundno
X_train = np.empty([roundnob,nopredict,repeat])
X_test = np.empty([roundno,nopredict,repeat])
y_train = np.empty([roundnob,repeat])
y_test = np.empty([roundno,repeat])
pred_train_rf = np.empty([roundnob,repeat])
pred_test_rf = np.empty([roundno,repeat])
misfit = np.empty([roundno,repeat])
inputsave = np.empty([len(lit4),nopredict+1,repeat])
model_rf = {}

# fit the data scaler
scaler.fit(lit4[predictors])

#save the data scaler
pickle.dump(scaler, open('scaler_%s.pkl'%timestrb,'wb'))

#scale the input data (N.B. ML algorithms perform better on scaled data) and assign to arrays
lit4[predictors] = scaler.transform(lit4[predictors])
X = lit4[predictors].values 
y = lit4[target_column].values
#%%  RANDOM FOREST REGRESSION MONTE CARLO

for i in range(0,repeat): 
  
    # split into randomly selected test and training datasets and save to arrays
    X_train[:,:,i], X_test[:,:,i], y_train[:,i], y_test[:,i] = train_test_split(X, y[:,0], test_size=frac)

    #define and fit RFR model - parameters selected using grid searching elsewhere
    model_rf[i] = RandomForestRegressor(bootstrap = False, max_depth = None, min_samples_leaf = 1, min_samples_split = 3, n_estimators=75, max_features='sqrt')
    model_rf[i].fit(X_train[:,:,i], y_train[:,i]) 
    
    #predict pressures for training and test datasets
    pred_train_rf[:,i] = model_rf[i].predict(X_train[:,:,i])
    pred_test_rf[:,i] = model_rf[i].predict(X_test[:,:,i])
    
    #print out R^2 fitting stats for test and training dataset
    print(r2_score(y_test[:,i], pred_test_rf[:,i]))

    #calculate misfit between predicted pressures in test dataset and measured pressures
    misfit[:,i] = pred_test_rf[:,i]-y_test[:,i]
    
    
#%% RANDOM FOREST REGRESSION LEAVE ONE OUT VALIDATION

loo = model_selection.LeaveOneOut()

#define RFR model for LOO
model_rf_loo = RandomForestRegressor(bootstrap = False,criterion='mse', max_depth = None, min_samples_leaf = 1, min_samples_split = 3, n_estimators=75, max_features='sqrt')

#Create empty arrays to populate
y_measure = np.empty(0)
y_pred = np.empty(0)
  
#LOO loop
for loop_number, (train, test) in enumerate(loo.split(X)):

    ## Make LOO training and testing arrays from full dataset
    training_X_array = X[train]
    training_y_array = y[train].reshape(-1, 1)
    X_test_array = X[test]
    y_actual_values = y[test]

    ## Fit RFR LOO models  
    model_rf_loo.fit(training_X_array, training_y_array[:,0])
    ## Predict pressure of unknown (the one left out)
    pred_test_rfloo = model_rf_loo.predict(X_test_array)      
   
    #append predicted and measured pressures to array
    y_pred = np.append(y_pred, pred_test_rfloo)
    y_measure = np.append(y_measure, y_actual_values[0][0])
    
#%% parameters for MC performance plots

width = 0.4 # the width of the bars 
for i in range(0,repeat):
    imp = model_rf[i].feature_importances_
    ind = np.arange(len(imp)) # the x locations for the groups
    
#save R^2 and RMSE stats for all training and test sets
r2scoretest = np.empty(repeat)
r2scoretrain = np.empty(repeat)
RMSE=np.empty(repeat)
for i in range(0,repeat):
    r2scoretest[i] = r2_score(y_test[:,i], pred_test_rf[:,i])
    r2scoretrain[i] = r2_score(y_train[:,i], pred_train_rf[:,i])
    RMSE[i] = math.sqrt(mean_squared_error(y_test[:,i], pred_test_rf[:,i]))
   
#calculate predicted pressure for each experimental composition
Xs = lit4[predictors].values   
pred_p = np.empty([len(lit4),repeat])
for i in range(0,repeat):
    pred_p[:,i] = model_rf[i].predict(Xs)

#calculate average pressure and 2 sigma for each input composition
pressures_pred = np.empty([len(Xs)])
pre_std_pred = np.empty([len(Xs)])
for i in range(0,len(Xs)):
    pressures_pred[i] = np.mean(pred_p[i,:])
    pre_std_pred[i] = 2 * np.std(pred_p[i,:])
    
#append calculated pressures to dataframe
lit4['pressure_predicted'] = pressures_pred
lit4['pressure_2sigma'] = pre_std_pred
lit4['pressure_LOOCV'] = y_pred
lit5['pressure_predicted_splits'] = pressures_pred
lit5['pressure_2sigma'] = pre_std_pred
lit5['pressure_LOOCV'] = y_pred

#write predicted pressures into file - alongside all input info
lit5.to_csv(r'out_lit_data.txt',sep='\t',index=False,header=True)

#%% make and save Figure 5

fig5, ((f51, f52), (f53, f54)) = plt.subplots(2,2, figsize=(14, 10))
width = 0.4 # the width of the bars 
for i in range(0,repeat):
    imp = model_rf[i].feature_importances_
    ind = np.arange(len(imp)) # the x locations for the groups
    f51.plot(ind, imp, width,alpha=0.2)
f51.set_xticks(ind+width/10)
f51.set_xticklabels(predictors, minor=False)
f51.set_title('Feature importance in RandomForest Classifier')
f51.set_ylabel('Relative importance')
f51.set_xlabel('feature') 

sns.distplot(r2scoretest, hist=True, kde=True, 
             bins=30, color = 'darkblue', 
             hist_kws={'edgecolor':'None'},
             kde_kws={'linewidth': 1.5},ax=f52,label="test R$^{2}$")

f52.set_xlabel("R$^2$ score")
f52.set_ylabel("count")
f52.set_xlim(0.68,0.95)

me_r2 = np.mean(r2scoretest)
me_rm = np.mean(RMSE)

for i in range(len(lit4)):
    f53.errorbar(lit4['pressure_predicted'][i],lit4['ExptP'][i],xerr=lit4['pressure_2sigma'][i], ecolor='k',elinewidth = 0.25 ,marker=lit4['mark'][i], markersize = 7, mec='k',mew=0.25,color=lit4['col'][i],linestyle='None',alpha=0.66)
f53.plot((0,30),(0,30),'k-')
f53.plot((0,30),(2.5,32.5),'k--',lw=0.5,zorder=1)
f53.plot((2.5,32.5),(0,30),'k--',lw=0.5,zorder=1)
f53.set_xlim(0,30)
f53.set_ylim(0,30)
f53.text(15, 28, 'mean $R^{2}$: %.2f, RMSE: %.3f'%(me_r2,me_rm) ,ha='center',va='center')
f53.set_ylabel('measured pressure (GPa)')
f53.set_xlabel('predicted pressure (GPa)')


loo_r2 = r2_score(y_pred,y_measure)
loo_rm = (mean_squared_error(y_pred,y_measure))**0.5

for i in range(len(lit4)):
    f54.scatter(y_pred[i],y_measure[i],marker=lit4['mark'][i], s=40, edgecolors='k', linewidths=0.25,color=lit4['col'][i],linestyle='None',alpha=0.66,zorder=2)
f54.plot((0,30),(0,30),'k-',zorder=1)
f54.set_xlim(0,30)
f54.set_ylim(0,30)
f54.plot((0,30),(2.5,32.5),'k--',lw=0.5,zorder=1)
f54.plot((2.5,32.5),(0,30),'k--',lw=0.5,zorder=1)

f54.text(15,28, '$R^{2}$: %.2f, RMSE: %.3f'%(loo_r2,loo_rm),ha='center',va='center')
f54.set_ylabel('measured pressure (GPa)')
f54.set_xlabel('predicted pressure (GPa)')

plt.show()
# fig5.savefig("fig5_%s.pdf"%timestr, bbox_inches='tight')

#%% Train RFR barometer using all input data - for inclusion prediction pressures

df_final = lit4.reset_index(drop=True)
Xfinal = df_final[predictors].values 
yfinal = df_final[target_column].values

final_rf = RandomForestRegressor(bootstrap = False, max_depth = None, min_samples_leaf = 1, min_samples_split = 3, n_estimators=75, max_features='sqrt')
final_rf.fit(Xfinal, yfinal[:,0]) 
final_pred_rf = final_rf.predict(Xfinal)
df_final['pressure_predicted_final'] = final_pred_rf

pkl_filename = "pickle_model_%s.pkl"%timestrb
with open(pkl_filename, 'wb') as file:
    pickle.dump(final_rf,file,-1)     


#%% read in inclusion compositions, filter for majoritic garnets and calculate pressures

incl = pd.read_csv('di_incs_040221.txt',sep="\t") # LITERATURE INCLUSION DATASET

namj = np.empty([len(incl)])
maj = np.empty([len(incl)])
tot_maj = np.empty([len(incl)])

for i in range(0,len(incl)):
    if (incl.Na[i]-incl.Ti[i])/2 >= 0:
        namj[i] = (incl.Na[i]-incl.Ti[i])/2
    else:
        namj[i] = 0

    if (incl.Si[i]-(2*namj[i])-3) >= 0:
        maj[i] = (incl.Si[i]-(2*namj[i])-3)
    else:
        maj[i] = 0
        
    tot_maj[i] = namj[i]+maj[i]

incl['tot_maj'] = tot_maj
    
inclusions = incl[incl['tot_maj'] >= 0.01]

df2 = inclusions.drop(columns=['tot_maj','Country','Mine'])

del incl, namj, maj, tot_maj

#apply scaling used on training data to inclusion compositions
df2[predictors] = scaler.transform(df2[predictors])

#turn pandas DFs into numpy arrays
diX = df2[predictors].values   



#%% CALCULATE INCLUSION PRESSURES with normally distributed random uncertainties added to compositions

df_noise = df
df_noise = df_noise.reset_index(drop=True)   
    
repeat = 100
# Create arrays that will be populated in RFR loop
nopredict = len(predictors)
pred_press_noise = np.empty([len(diX),repeat])

df2b = inclusions.drop(columns=['tot_maj','Country','Mine'])
diXs = np.empty([len(df2b),nopredict,repeat])

noise_out = np.empty([repeat])
for i in range(0,repeat):   
    testinc = df2b/df2b
    testinc['Si'] = testinc['Si'] + np.random.normal(0,1,size=testinc['Si'].shape) / 500
    testinc['Ti'] = testinc['Ti'] + np.random.normal(0,1,size=testinc['Ti'].shape) / 33
    testinc['Al'] = testinc['Al'] + np.random.normal(0,1,size=testinc['Al'].shape) / 100
    testinc['Cr'] = testinc['Cr'] + np.random.normal(0,1,size=testinc['Cr'].shape) / 20
    testinc['Fet'] = testinc['Fet'] + np.random.normal(0,1,size=testinc['Fet'].shape) / 33
    testinc['Mn'] = testinc['Mn'] + np.random.normal(0,1,size=testinc['Mn'].shape) / 10
    testinc['Mg'] = testinc['Mg'] + np.random.normal(0,1,size=testinc['Mg'].shape) / 33
    testinc['Ca'] = testinc['Ca'] + np.random.normal(0,1,size=testinc['Ca'].shape) / 100
    testinc['Na'] = testinc['Na'] + np.random.normal(0,1,size=testinc['Na'].shape) / 33
    testinc = testinc.fillna(1)
    df2b_temp = testinc * df2b
    df2b_temp[predictors] = scaler.transform(df2b_temp[predictors])
    diX_temp = df2b_temp[predictors].values   
    diXs[:,:,i] = diX_temp[:,:]
    
    #predict pressures for training dataset
    pred_press_noise[:,i] = final_rf.predict(diX_temp)

#append all calculated inclusion pressures    
data_noise = pred_press_noise[:,0]
for i in range(1,repeat):
    data_noise = np.append(data_noise,pred_press_noise[:,i])

#inclusion pressures mean, stddev, min and max
pressures_preddi = np.empty([len(pred_press_noise)])
pre_std_preddi = np.empty([len(pred_press_noise)])
pressures_maxdi = np.empty([len(pred_press_noise)])
pressures_mindi = np.empty([len(pred_press_noise)])
for i in range(0,len(pred_press_noise)):
    pressures_preddi[i] = np.mean(pred_press_noise[i,:])
    pre_std_preddi[i] = 2 * np.std(pred_press_noise[i,:])
    pressures_maxdi[i] = np.max(pred_press_noise[i,:])
    pressures_mindi[i] = np.min(pred_press_noise[i,:])

inclusions['pressure_best'] = pressures_preddi
inclusions['Pbest_error'] = pre_std_preddi

#save calculated inclusion pressures
inclusions.to_csv(r'diout.txt',sep='\t',index=False,header=True)


#%% GLOBAL DI-INCLUSION DISTRIBUTION figure

xs = np.linspace(5, 26, 1000)
N = np.arange(1, 5) 
#FIGURE OF INC PREESURES - HIST, PDE and GMM - good models only
        
datafinalb = data_noise.reshape(-1, 1)
models_final = [None for i in range(len(N))]


for i in range(len(N)):
    models_final[i] = GaussianMixture(N[i]).fit(datafinalb)

# compute the AIC and the BIC
AIC = [m.aic(datafinalb) for m in models_final]
BIC = [m.bic(datafinalb) for m in models_final]


Mfinal_best = models_final[np.argmin(AIC)]

logprobfinal = Mfinal_best.score_samples(xs.reshape(-1, 1))
responsibilitiesfinal = Mfinal_best.predict_proba(xs.reshape(-1, 1))
pdffinal = np.exp(logprobfinal)
pdf_individualfinal = responsibilitiesfinal * pdffinal[:, np.newaxis]

fig_incs = plt.figure(figsize=(7,5))  
sns.distplot(data_noise, hist=True, kde=True, 
             bins=25, color = 'darkblue', 
             hist_kws={'edgecolor':'None','alpha':0.2,'zorder':2},
             kde_kws={'linewidth': 2, 'zorder':4}, label="Final RF barometer")


plt.plot(xs, pdffinal, '-', color='k', lw=2,zorder=5)
plt.plot(xs, pdf_individualfinal, 'k--',lw=0.5,zorder=2)

plt.xlim(3,27)
plt.xlabel("predicted pressure (GPa)")
plt.ylabel("probability (normalised count)")
plt.title("100 MC iterations with random uncertainties")
plt.legend(loc="upper right", fontsize = 'medium')
plt.xlabel("predicted pressure (GPa)")
plt.ylabel("probability (normalised count)")

plt.show()
# fig_incs.savefig("fig_incs_%s.pdf"%timestr, bbox_inches='tight')

#%%

import rpy2.robjects as robjects

from rpy2.robjects.packages import importr
from rpy2.robjects import r, pandas2ri

import rpy2.robjects as ro
from rpy2.robjects import pandas2ri
pandas2ri.activate()

randomForest = importr('randomForest')
importr('caTools')
tidyverse = importr('tidyverse')
caret = importr('caret')
Metrics = importr('Metrics')
e1071 = importr('e1071')

#%%

lit_maj_nat = ro.conversion.py2rpy(df)
